/* Copyright (C) 2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Volker Friese [committer], Sebastian Heinemann, Felix Weiglhofer */


#include "Unpack.h"

#include <chrono>

#include <xpu/host.h>

#include "compat/Algorithm.h"
#include "log.hpp"

using namespace std;
using fles::Subsystem;

namespace cbm::algo
{
  // -----   Execution   -------------------------------------------------------
  Unpack::resultType Unpack::operator()(const fles::Timeslice* timeslice)
  {
    xpu::push_timer("Unpack");

    // --- Output data
    resultType result          = {};
    CbmDigiTimeslice& digiTs   = result.first;
    UnpackMonitorData& monitor = result.second;

    if (DetectorEnabled(Subsystem::STS)) {
      monitor.fNumBytesInSts +=
        ParallelMsLoop(Subsystem::STS, monitor, digiTs.fData.fSts.fDigis, monitor.fSts, *timeslice, fAlgoSts, 0x20);
    }

    if (DetectorEnabled(Subsystem::TOF)) {
      monitor.fNumBytesInTof +=
        ParallelMsLoop(Subsystem::TOF, monitor, digiTs.fData.fTof.fDigis, monitor.fTof, *timeslice, fAlgoTof, 0x00);
    }

    if (DetectorEnabled(Subsystem::BMON)) {
      monitor.fNumBytesInBmon +=
        ParallelMsLoop(Subsystem::BMON, monitor, digiTs.fData.fT0.fDigis, monitor.fBmon, *timeslice, fAlgoBmon, 0x00);
    }

    if (DetectorEnabled(Subsystem::MUCH)) {
      monitor.fNumBytesInMuch +=
        ParallelMsLoop(Subsystem::MUCH, monitor, digiTs.fData.fMuch.fDigis, monitor.fMuch, *timeslice, fAlgoMuch, 0x20);
    }

    // ---  Component loop
    for (uint64_t comp = 0; comp < timeslice->num_components(); comp++) {

      // System ID of current component
      const auto subsystem = static_cast<Subsystem>(timeslice->descriptor(comp, 0).sys_id);

      if (!DetectorEnabled(subsystem)) continue;

      xpu::scoped_timer t1(fles::to_string(subsystem));
      xpu::t_add_bytes(timeslice->size_component(comp));

      // Equipment ID of current component
      const uint16_t equipmentId = timeslice->descriptor(comp, 0).eq_id;

      // The current algorithms work for the format versions hard-coded as parameters to MsLoop() below.
      // Other versions are not yet supported.
      // In the future, different data formats will be supported by instantiating different
      // algorithms depending on the version.

      // if (subsystem == Subsystem::STS) {
      //   MsLoop(timeslice, fAlgoSts, comp, equipmentId, &digiTs.fData.fSts.fDigis, monitor, &monitor.fSts, 0x20);
      // }
      // if (subsystem == Subsystem::MUCH) {
      //   MsLoop(timeslice, fAlgoMuch, comp, equipmentId, &digiTs.fData.fMuch.fDigis, monitor, &monitor.fMuch, 0x20);
      // }
      // if (subsystem == Subsystem::TOF) {
      //   MsLoop(timeslice, fAlgoTof, comp, equipmentId, &digiTs.fData.fTof.fDigis, monitor, &monitor.fTof, 0x00);
      // }
      // if (subsystem == Subsystem::BMON) {
      //   MsLoop(timeslice, fAlgoBmon, comp, equipmentId, &digiTs.fData.fT0.fDigis, monitor, &monitor.fBmon, 0x00);
      // }
      if (subsystem == Subsystem::TRD) {
        monitor.fNumBytesInTrd +=
          MsLoop(timeslice, fAlgoTrd, comp, equipmentId, &digiTs.fData.fTrd.fDigis, monitor, &monitor.fTrd, 0x01);
      }
      if (subsystem == Subsystem::TRD2D) {
        monitor.fNumBytesInTrd2d +=
          MsLoop(timeslice, fAlgoTrd2d, comp, equipmentId, &digiTs.fData.fTrd2d.fDigis, monitor, &monitor.fTrd2d, 0x02);
      }
      if (subsystem == Subsystem::RICH) {
        monitor.fNumBytesInRich +=
          MsLoop(timeslice, fAlgoRich, comp, equipmentId, &digiTs.fData.fRich.fDigis, monitor, &monitor.fRich, 0x03);
      }
    }  //# component

    // --- Sorting of output digis. Is required by both digi trigger and event builder.

    xpu::push_timer("Sort");
    auto& digiData = digiTs.fData;
    DoSort(digiData.fSts.fDigis);
    DoSort(digiData.fMuch.fDigis);
    DoSort(digiData.fTof.fDigis);
    DoSort(digiData.fT0.fDigis);
    DoSort(digiData.fTrd.fDigis);
    DoSort(digiData.fTrd2d.fDigis);
    DoSort(digiData.fRich.fDigis);
    xpu::pop_timer();

    xpu::timings timings = xpu::pop_timer();

    monitor.fTimeUnpack = timings.wall();

    return result;
  }
  // ----------------------------------------------------------------------------


  // ----------------------------------------------------------------------------
  std::pair<size_t, size_t> Unpack::ParallelInit(const fles::Timeslice& ts, Subsystem subsystem,
                                                 gsl::span<const uint16_t> legalEqIds, uint8_t sys_ver,
                                                 UnpackMonitorData& monitor, std::vector<u16>& msEqIds,
                                                 std::vector<fles::MicrosliceDescriptor>& msDesc,
                                                 std::vector<const u8*>& msContent)
  {
    size_t numMs     = 0;
    size_t sizeBytes = 0;

    for (uint64_t comp = 0; comp < ts.num_components(); comp++) {
      auto this_subsystem = static_cast<Subsystem>(ts.descriptor(comp, 0).sys_id);

      if (this_subsystem == subsystem) {
        const u64 numMsInComp = ts.num_microslices(comp);
        const u16 componentId = ts.descriptor(comp, 0).eq_id;

        if (ts.descriptor(comp, 0).sys_ver != sys_ver) {
          monitor.fNumErrInvalidSysVer++;
          continue;
        }

        if (std::find(legalEqIds.begin(), legalEqIds.end(), componentId) == legalEqIds.end()) {
          monitor.fNumErrInvalidEqId++;
          continue;
        }

        sizeBytes += ts.size_component(comp);
        numMs += numMsInComp;
        for (u64 mslice = 0; mslice < numMsInComp; mslice++) {
          uint64_t msByteSize = ts.descriptor(comp, mslice).size;
          msEqIds.push_back(componentId);
          msDesc.push_back(ts.descriptor(comp, mslice));
          msContent.push_back(ts.content(comp, mslice));
        }
      }
    }

    return {numMs, sizeBytes};
  }
  // ----------------------------------------------------------------------------


  // ----------------------------------------------------------------------------
  template<class Digi, class UnpackAlgo, class Monitor>
  size_t Unpack::ParallelMsLoop(const Subsystem subsystem, UnpackMonitorData& genericMonitor,
                                std::vector<Digi>& digisOut, std::vector<Monitor>& monitorOut,
                                const fles::Timeslice& ts, const std::map<u16, UnpackAlgo>& algos, u8 sys_ver)
  {
    xpu::scoped_timer t_(fles::to_string(subsystem));

    std::vector<u16> msEqIds;                        // equipment ids of microslices
    std::vector<fles::MicrosliceDescriptor> msDesc;  // microslice descriptors
    std::vector<const u8*> msContent;                // pointer to microslice content
    auto legalEqIds = GetEqIds(algos);
    auto [numMs, sizeBytes] =
      ParallelInit(ts, subsystem, gsl::make_span(legalEqIds), sys_ver, genericMonitor, msEqIds, msDesc, msContent);
    std::vector<std::vector<Digi>> msDigis(numMs);  // unpacked digis per microslice
    std::vector<Monitor> monitor(numMs);            // unpacking monitoring data per microslice

    xpu::t_add_bytes(sizeBytes);
    genericMonitor.fNumBytes += sizeBytes;

    xpu::push_timer("Unpack");
#pragma omp parallel for schedule(dynamic)
    for (size_t i = 0; i < numMs; i++) {
      auto result = algos.at(msEqIds[i])(msContent[i], msDesc[i], ts.start_time());
      msDigis[i]  = std::move(result.first);
      monitor[i]  = std::move(result.second);
    }
    xpu::pop_timer();

    size_t nDigisTotal = 0;
    for (const auto& digis : msDigis) {
      nDigisTotal += digis.size();
    }

    xpu::push_timer("Resize");
    digisOut.resize(nDigisTotal);
    xpu::pop_timer();

    xpu::push_timer("Merge");
#pragma omp parallel for schedule(dynamic)
    for (unsigned int i = 0; i < numMs; i++) {
      unsigned int offset = 0;
      for (unsigned int x = 0; x < i; x++)
        offset += msDigis[x].size();
      std::copy(msDigis[i].begin(), msDigis[i].end(), digisOut.begin() + offset);
    }
    xpu::pop_timer();

    monitorOut = std::move(monitor);

    return sizeBytes;
  }
  // ----------------------------------------------------------------------------


  // ----------------- Microslice loop ------------------------------------------
  template<class Digi, class UnpackAlgo, class MonitorData>
  size_t Unpack::MsLoop(const fles::Timeslice* timeslice, std::map<uint16_t, UnpackAlgo>& algoMap, const uint64_t comp,
                        const uint16_t eqId, std::vector<Digi>* digis, UnpackMonitorData& monitor,
                        std::vector<MonitorData>* monitorMs, uint8_t sys_ver)
  {
    // --- Component log
    size_t numBytesInComp = 0;
    size_t numDigisInComp = 0;

    // For profiling
    const auto starttime = std::chrono::high_resolution_clock::now();

    // Get Unpacker
    const auto algoIt = algoMap.find(eqId);
    if (algoIt == algoMap.end()) {
      monitor.fNumErrInvalidEqId++;
      return 0;
    }
    UnpackAlgo& algo = algoIt->second;

    if (timeslice->descriptor(comp, 0).sys_ver != sys_ver) {
      monitor.fNumErrInvalidSysVer++;
      return 0;
    }

    const uint64_t numMsInComp = timeslice->num_microslices(comp);

    for (uint64_t mslice = 0; mslice < numMsInComp; mslice++) {
      const auto msDescriptor = timeslice->descriptor(comp, mslice);
      const auto msContent    = timeslice->content(comp, mslice);
      auto result             = algo(msContent, msDescriptor, timeslice->start_time());
      L_(debug) << "Unpack::MsLoop(): Component " << comp << ", microslice " << mslice << ", digis "
                << result.first.size() << ", " << result.second.print();
      numBytesInComp += msDescriptor.size;
      numDigisInComp += result.first.size();
      digis->insert(digis->end(), result.first.begin(), result.first.end());
      monitorMs->push_back(result.second);
    }
    // Get elapsed time
    const auto endtime  = std::chrono::high_resolution_clock::now();
    const auto duration = std::chrono::duration_cast<std::chrono::microseconds>(endtime - starttime);

    L_(debug) << "Unpack(): Component " << comp << ", subsystem "
              << fles::to_string(static_cast<Subsystem>(timeslice->descriptor(comp, 0).sys_id)) << ", microslices "
              << numMsInComp << " input size " << numBytesInComp << " bytes,"
              << " digis " << numDigisInComp << ", CPU time " << duration.count() / 1000. << " ms";

    monitor.fNumMs += numMsInComp;
    monitor.fNumBytes += numBytesInComp;
    monitor.fNumDigis += numDigisInComp;
    monitor.fNumCompUsed++;

    return numBytesInComp;
  }
  // ----------------------------------------------------------------------------


  // -----   Initialisation   ---------------------------------------------------
  void Unpack::Init(std::vector<Subsystem> subIds)
  {
    fSubsystems = subIds;

    // --- Common parameters for all components for STS
    uint32_t numChansPerAsicSts   = 128;  // R/O channels per ASIC for STS
    uint32_t numAsicsPerModuleSts = 16;   // Number of ASICs per module for STS

    // Create one algorithm per component for STS and configure it with parameters
    auto equipIdsSts = fStsConfig.GetEquipmentIds();
    for (auto& equip : equipIdsSts) {
      std::unique_ptr<UnpackStsPar> par(new UnpackStsPar());
      par->fNumChansPerAsic   = numChansPerAsicSts;
      par->fNumAsicsPerModule = numAsicsPerModuleSts;
      const size_t numElinks  = fStsConfig.GetNumElinks(equip);
      for (size_t elink = 0; elink < numElinks; elink++) {
        UnpackStsElinkPar elinkPar;
        auto mapEntry        = fStsConfig.Map(equip, elink);
        elinkPar.fAddress    = mapEntry.first;   // Module address for this elink
        elinkPar.fAsicNr     = mapEntry.second;  // ASIC number within module
        elinkPar.fTimeOffset = fSystemTimeOffset[Subsystem::STS];
        elinkPar.fAdcMinCut  = fStsConfig.AdcCutMap(equip, elink);
        elinkPar.fAdcOffset  = 1.;
        elinkPar.fAdcGain    = 1.;
        if (fApplyWalkCorrection) elinkPar.fWalk = fStsConfig.WalkMap(elinkPar.fAddress, elinkPar.fAsicNr);
        elinkPar.fChanMask = fStsConfig.MaskMap(equip, elink);
        // TODO: Add parameters for time and ADC calibration
        par->fElinkParams.push_back(elinkPar);
      }
      fAlgoSts[equip].SetParams(std::move(par));
      L_(debug) << "--- Configured equipment " << equip << " with " << numElinks << " elinks";
    }  //# equipments

    // Create one algorithm per component for MUCH and configure it with parameters
    auto equipIdsMuch = fMuchConfig.GetEquipmentIds();
    for (auto& equip : equipIdsMuch) {
      std::unique_ptr<UnpackMuchPar> par(new UnpackMuchPar());
      const size_t numElinks = fMuchConfig.GetNumElinks(equip);
      for (size_t elink = 0; elink < numElinks; elink++) {
        UnpackMuchElinkPar elinkPar;
        elinkPar.fAddress    = fMuchConfig.Map(equip, elink);  // Vector of MUCH addresses for this elink
        elinkPar.fTimeOffset = fSystemTimeOffset[Subsystem::MUCH];
        elinkPar.fChanMask   = fMuchConfig.MaskMap(equip, elink);
        par->fElinkParams.push_back(elinkPar);
      }
      fAlgoMuch[equip].SetParams(std::move(par));
      L_(debug) << "--- Configured equipment " << equip << " with " << numElinks << " elinks";
    }

    // Create one algorithm per component for TOF and configure it with parameters
    auto equipIdsTof = fTofConfig.GetEquipmentIds();
    for (auto& equip : equipIdsTof) {
      std::unique_ptr<UnpackTofPar> par(new UnpackTofPar());
      const size_t numElinks = fTofConfig.GetNumElinks(equip);
      for (size_t elink = 0; elink < numElinks; elink++) {
        UnpackTofElinkPar elinkPar;
        elinkPar.fChannelUId = fTofConfig.Map(equip, elink);  // Vector of TOF addresses for this elink
        elinkPar.fTimeOffset = fSystemTimeOffset[Subsystem::TOF];
        par->fElinkParams.push_back(elinkPar);
      }
      fAlgoTof[equip].SetParams(std::move(par));
      L_(debug) << "--- Configured equipment " << equip << " with " << numElinks << " elinks";
    }

    // Create one algorithm per component for T0 and configure it with parameters
    auto equipIdsBmon = fBmonConfig.GetEquipmentIds();
    for (auto& equip : equipIdsBmon) {
      std::unique_ptr<UnpackBmonPar> par(new UnpackBmonPar());
      const size_t numElinks = fBmonConfig.GetNumElinks(equip);
      for (size_t elink = 0; elink < numElinks; elink++) {
        UnpackBmonElinkPar elinkPar;
        elinkPar.fChannelUId = fBmonConfig.Map(equip, elink);  // Vector of T0 addresses for this elink
        elinkPar.fTimeOffset = fSystemTimeOffset[Subsystem::BMON];
        par->fElinkParams.push_back(elinkPar);
      }
      fAlgoBmon[equip].SetParams(std::move(par));
      L_(debug) << "--- Configured equipment " << equip << " with " << numElinks << " elinks";
    }

    // Create one algorithm per component and configure it with parameters
    auto equipIdsRich = fRichConfig.GetEquipmentIds();
    for (auto& equip : equipIdsRich) {
      std::unique_ptr<UnpackRichPar> par(new UnpackRichPar());
      std::map<uint32_t, std::vector<double>> compMap = fRichConfig.Map(equip);
      for (auto const& val : compMap) {
        uint32_t address                       = val.first;
        par->fElinkParams[address].fToTshift   = val.second;
        par->fElinkParams[address].fTimeOffset = fSystemTimeOffset[Subsystem::RICH];
      }
      fAlgoRich[equip].SetParams(std::move(par));
      L_(info) << "--- Configured equipment " << equip << " with " << fRichConfig.GetNumElinks(equip) << " elinks";
    }

    // Create one algorithm per component for TRD and configure it with parameters
    auto equipIdsTrd = fTrdConfig.GetEquipmentIds();
    for (auto& equip : equipIdsTrd) {

      std::unique_ptr<UnpackTrdPar> par(new UnpackTrdPar());
      const size_t numCrobs = fTrdConfig.GetNumCrobs(equip);

      for (size_t crob = 0; crob < numCrobs; crob++) {
        UnpackTrdCrobPar crobPar;
        const size_t numElinks = fTrdConfig.GetNumElinks(equip, crob);

        for (size_t elink = 0; elink < numElinks; elink++) {
          UnpackTrdElinkPar elinkPar;
          auto addresses        = fTrdConfig.Map(equip, crob, elink);
          elinkPar.fAddress     = addresses.first;   // Asic address for this elink
          elinkPar.fChanAddress = addresses.second;  // Channel addresses for this elink
          elinkPar.fTimeOffset  = fSystemTimeOffset[Subsystem::TRD];
          crobPar.fElinkParams.push_back(elinkPar);
        }
        par->fCrobParams.push_back(crobPar);
      }
      fAlgoTrd[equip].SetParams(std::move(par));
      L_(debug) << "--- Configured equipment " << equip << " with " << numCrobs << " crobs";
    }

    // Create one algorithm per component for TRD2D and configure it with parameters
    auto equipIdsTrd2d = fTrd2dConfig.GetEquipmentIds();
    for (auto& equip : equipIdsTrd2d) {

      std::unique_ptr<UnpackTrd2dPar> par(new UnpackTrd2dPar());
      const size_t numAsics = fTrd2dConfig.GetNumAsics(equip);

      for (size_t asic = 0; asic < numAsics; asic++) {
        UnpackTrd2dAsicPar asicPar;
        const size_t numChans = fTrd2dConfig.GetNumChans(equip, asic);

        for (size_t chan = 0; chan < numChans; chan++) {
          UnpackTrd2dChannelPar chanPar;
          auto pars           = fTrd2dConfig.ChanMap(equip, asic, chan);
          chanPar.fPadAddress = std::get<0>(pars);  // Pad address for channel
          chanPar.fMask       = std::get<1>(pars);  // Flag channel mask
          chanPar.fDaqOffset  = std::get<2>(pars);  // Time calibration parameter
          asicPar.fChanParams.push_back(chanPar);
        }
        auto comppars          = fTrd2dConfig.CompMap(equip);
        par->fSystemTimeOffset = fSystemTimeOffset[Subsystem::TRD2D];
        par->fModId            = comppars.first;
        par->fCrobId           = comppars.second;
        par->fAsicParams.push_back(asicPar);
      }
      fAlgoTrd2d[equip].SetParams(std::move(par));
      L_(debug) << "--- Configured equipment " << equip << " with " << numAsics << " asics";
    }

    L_(info) << "--- Configured " << fAlgoSts.size()
             << " unpacker algorithms for STS. (Walk correction = " << fApplyWalkCorrection << ")";
    L_(debug) << "Readout map:" << fStsConfig.PrintReadoutMap();
    L_(info) << "--- Configured " << fAlgoMuch.size() << " unpacker algorithms for MUCH.";
    L_(info) << "--- Configured " << fAlgoRich.size() << " unpacker algorithms for RICH.";
    L_(debug) << "Readout map:" << fRichConfig.PrintReadoutMap();
    L_(info) << "--- Configured " << fAlgoTof.size() << " unpacker algorithms for TOF.";
    L_(info) << "--- Configured " << fAlgoTrd.size() << " unpacker algorithms for TRD.";
    L_(info) << "--- Configured " << fAlgoTrd2d.size() << " unpacker algorithms for TRD2D.";
    L_(info) << "--- Configured " << fAlgoBmon.size() << " unpacker algorithms for T0.";
    L_(info) << "==================================================";
  }
  // ----------------------------------------------------------------------------


  // ----------------------------------------------------------------------------
  template<class Digi>
  void Unpack::DoSort(std::vector<Digi>& digis)
  {
    xpu::t_add_bytes(digis.size() * sizeof(Digi));  // Add bytes to timer, assumes xpu::timers are started in operator()
    Sort(digis.begin(), digis.end(),
         [](const Digi& digi1, const Digi& digi2) { return digi1.GetTime() < digi2.GetTime(); });
  }


} /* namespace cbm::algo */
