/* Copyright (C) 2023 Physikalisches Institut, Eberhard Karls Universitaet Tuebingen, Tuebingen
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Lukas Chlad [committer] */

#include "CbmFsdDetectorMapManager.h"

#include "CbmFsdDetectorData.h"  // for CbmFsdModuleData

#include <Logger.h>  // for LOG, Logger

#include <TGeoBBox.h>     // for TGeoBBox
#include <TGeoManager.h>  // for TGeoManager, gGeoManager
#include <TGeoMatrix.h>   // for TGeoMatrix
#include <TGeoNode.h>     // for TGeoIterator, TGeoNode
#include <TGeoVolume.h>   // for TGeoVolume
#include <TRandom.h>      // for TRandom, gRandom

#include <string>   // for operator<, stoul
#include <utility>  // for pair

#include <stddef.h>  // for size_t

using namespace std;

CbmFsdDetectorMapManager::CbmFsdDetectorMapManager() : fModulePathToIdMap(), fModuleIdToDataMap(), fModuleIds()
{
  Init();
}

CbmFsdDetectorMapManager::~CbmFsdDetectorMapManager() {}

void CbmFsdDetectorMapManager::Init()
{

  fModulePathToIdMap.clear();
  fModuleIdToDataMap.clear();
  fModuleIds.clear();

  Int_t currentModuleId = 0;

  TGeoIterator geoIterator(gGeoManager->GetTopNode()->GetVolume());

  TGeoNode* curNode;
  TString branchStr("fsd_");  // stored in path of nodes, fsd_geoTag_0
  TString moduleStr("fsdmodule");
  TString scintStr("scintillator");
  geoIterator.Reset();  // safety to reset to "cave" befor the loop starts
  while ((curNode = geoIterator())) {
    TString nodePath;
    geoIterator.GetPath(nodePath);
    if (!nodePath.Contains(branchStr)) {
      geoIterator.Skip();  // skip current branch when it is not FSD => should speed up
      continue;            // don't do anything for this branch
    }

    TString nodeName(curNode->GetName());
    // FSD detector -> all modules names contain "fsdmodule"
    if (TString(curNode->GetVolume()->GetName()).Contains(moduleStr)) {

      const TGeoMatrix* curMatrix = geoIterator.GetCurrentMatrix();
      const Double_t* curNodeTr   = curMatrix->GetTranslation();
      string path                 = string(nodePath.Data());

      TGeoVolume* curScintillator = nullptr;
      for (int idn = 0; idn < curNode->GetNdaughters(); idn++) {
        TGeoNode* daughterNode = curNode->GetDaughter(idn);
        if (TString(daughterNode->GetVolume()->GetName()).Contains(scintStr)) {
          curScintillator = daughterNode->GetVolume();
          break;
        }
      }
      const TGeoBBox* shape = (const TGeoBBox*) (curScintillator->GetShape());

      fModulePathToIdMap.insert(pair<string, Int_t>(path, currentModuleId));
      CbmFsdModuleData* moduleData = new CbmFsdModuleData();
      moduleData->fX               = curNodeTr[0];
      moduleData->fY               = curNodeTr[1];
      moduleData->fZ               = curNodeTr[2];
      moduleData->dX               = shape->GetDX();
      moduleData->dY               = shape->GetDY();
      moduleData->dZ               = shape->GetDZ();
      moduleData->fId              = currentModuleId;
      fModuleIdToDataMap.insert(pair<Int_t, CbmFsdModuleData*>(moduleData->fId, moduleData));
      fModuleIds.push_back(currentModuleId);
      currentModuleId++;
    }
  }

  LOG(info) << "CbmFsdDetectorMapManager is initialized";
  LOG(info) << "fModulePathToIdMap.size() = " << fModulePathToIdMap.size();
  LOG(info) << "fModuleIdToDataMap.size() = " << fModuleIdToDataMap.size();
}

Int_t CbmFsdDetectorMapManager::GetModuleIdByPath(const string& path)
{
  std::map<string, Int_t>::iterator it;
  it = fModulePathToIdMap.find(path);
  if (it == fModulePathToIdMap.end()) return -1;
  return it->second;
}


CbmFsdModuleData* CbmFsdDetectorMapManager::GetModuleDataById(Int_t modId)
{
  std::map<Int_t, CbmFsdModuleData*>::iterator it;
  it = fModuleIdToDataMap.find(modId);
  if (it == fModuleIdToDataMap.end()) return nullptr;
  return it->second;
}

vector<Int_t> CbmFsdDetectorMapManager::GetModuleIds() { return fModuleIds; }
