/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   CbmQaIO.h
/// @brief  Module for ROOT objects IO interface (header)
/// @author S.Zharko <s.zharko@gsi.de>
/// @since   29.03.2023

#ifndef CbmQaIO_h
#define CbmQaIO_h 1

#include "CbmQaTable.h"

#include "Logger.h"

#include "TCanvas.h"
#include "TEfficiency.h"
#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TObject.h"
#include "TParameter.h"
#include "TProfile.h"
#include "TProfile2D.h"
#include "TProfile3D.h"
#include "TROOT.h"
#include "TString.h"

#include <vector>

class TFile;

/// @brief  ROOT object IO interface for QA
///
/// The class provides interface to write ROOT object into resulted files
class CbmQaIO {
public:
  using ObjList_t = std::vector<std::pair<TObject*, TString>>;

  enum class EStoringMode
  {
    kSAMEDIR,  ///< Objects will be stored to root directory
    kSUBDIR    ///< Objects will be stored to corresponding subdirectory
  };

  /// @brief Constructor
  /// @param prefixName    Name of the unique prefix
  /// @param pParentFolder Pointer to parent folder
  CbmQaIO(const char* prefixName, std::shared_ptr<ObjList_t> pObjList = nullptr);

  /// @brief Destructor
  virtual ~CbmQaIO();

  /// @brief Copy constructor
  CbmQaIO(const CbmQaIO&) = delete;

  /// @brief Move constructor
  CbmQaIO(CbmQaIO&&) = delete;

  /// @brief Copy assignment operator
  CbmQaIO& operator=(const CbmQaIO&) = delete;

  /// @brief Move assignment operator
  CbmQaIO& operator=(CbmQaIO&&) = delete;

  /// @brief Creates, initializes and registers a canvas
  /// @tparam  T     Type of the canvas: TCanvas or CbmQaCanvas (\note T should not be a pointer)
  /// @param   name  Name of canvas
  /// @param   args  The rest of the arguments, which will be passed to the canvas constructor
  template<typename T, typename... Args>
  T* MakeCanvas(const char* name, Args... args);

  /// @brief Creates, initializes and registers an efficiency
  /// @tparam  T     Type of the canvas: either TProfile, TProfile1D or TEfficiency (\note T should not be a pointer)
  /// @param   name  Name of canvas
  /// @param   args  The rest of the arguments, which will be passed to the efficiency constructor
  template<typename T, typename... Args>
  T* MakeEfficiency(const char* name, Args... args);

  /// @brief Creates, initializes and registers a histogram
  /// @tparam  T     Type of the histogram (\note T should not be a pointer)
  /// @param   name  Name of histogram
  /// @param   args  The rest of the arguments, which will be passed to the histogram constructor
  template<typename T, typename... Args>
  T* MakeHisto(const char* name, Args... args);

  /// @brief Creates, initializes and registers a table
  /// @param  name  Name of the table
  /// @param  args  The rest of the arguments, which will be passed to the table constructor
  template<typename... Args>
  CbmQaTable* MakeTable(const char* name, Args... args);

  /// @brief Creates a ROOT object
  /// @param name    Name of the object
  /// @param args... Arguments passed to the object constructor
  template<typename T, typename... Args>
  T* MakeObject(const char* name, Args... args);

  /// @brief Sets a common root path to the objects in the output file
  /// @param path  A path to the object
  void SetRootFolderName(const char* path) { fsRootFolderName = path; }

protected:
  /// @brief Applies properties on the histogram created with MakeHisto funciton
  /// @param pHist  Pointer to histogram
  virtual void SetHistoProperties(TH1* /*pHits*/);

  /// @brief Writes objects into file
  /// @param pOutFile Pointer to output ROOT file
  void WriteToFile(TFile* pOutFile) const;

  TString fsRootFolderName = "";  ///< Name of root folder
  TString fsPrefix         = "";  ///< Unique prefix for all writeable root

  EStoringMode fStoringMode             = EStoringMode::kSUBDIR;  ///< Objects storing mode
  std::shared_ptr<ObjList_t> fpvObjList = nullptr;                ///< List of registered ROOT objects
};


// ---------------------------------------------------------------------------------------------------------------------
//
template<typename T, typename... Args>
T* CbmQaIO::MakeCanvas(const char* nameBase, Args... args)
{
  TString sFullName = EStoringMode::kSUBDIR == fStoringMode ? Form("canvases/%s", nameBase) : nameBase;
  auto* pCanv       = MakeObject<T>(sFullName, args...);
  return pCanv;
}

// ---------------------------------------------------------------------------------------------------------------------
//
template<typename T, typename... Args>
T* CbmQaIO::MakeEfficiency(const char* nameBase, Args... args)
{
  TString sFullName = EStoringMode::kSUBDIR == fStoringMode ? Form("efficiencies/%s", nameBase) : nameBase;
  auto* pEff        = MakeObject<T>(sFullName, args...);
  return pEff;
}

// ---------------------------------------------------------------------------------------------------------------------
//
template<typename T, typename... Args>
T* CbmQaIO::MakeHisto(const char* nameBase, Args... args)
{
  TString sFullName = EStoringMode::kSUBDIR == fStoringMode ? Form("histograms/%s", nameBase) : nameBase;
  auto* pHist       = MakeObject<T>(sFullName, args...);
  pHist->SetDirectory(0);
  SetHistoProperties(pHist);
  return pHist;
}

// ---------------------------------------------------------------------------------------------------------------------
//
template<typename T, typename... Args>
T* CbmQaIO::MakeObject(const char* name, Args... args)
{
  // Resolve directory name and object name
  TString sObjName   = name;
  auto iLastSlashPos = static_cast<int>(sObjName.Last('/'));
  TString sDirName   = sObjName(0, iLastSlashPos);
  sObjName           = sObjName(iLastSlashPos + 1, sObjName.Length() - iLastSlashPos);

  // Add unique prefix to the object name
  sObjName = fsPrefix + "_" + sObjName;

  // Add parent directory to the folder
  if (fsRootFolderName.Length() != 0) { sDirName = fsRootFolderName + "/" + sDirName; }

  // Check, if an object with the same name already exists
  if (gROOT->FindObject(sObjName)) {
    LOG(warn) << fsRootFolderName << ": A previously created object \"" << sObjName << "\" will be deleted";
    auto* pObj = static_cast<T*>(gROOT->FindObject(sObjName));
    delete pObj;
  }

  // Create a new object
  T* pObj = new T(sObjName, args...);

  // Register the object into an object list
  fpvObjList->push_back(std::make_pair(static_cast<TObject*>(pObj), sDirName));
  return pObj;
}

// ---------------------------------------------------------------------------------------------------------------------
//
template<typename... Args>
CbmQaTable* CbmQaIO::MakeTable(const char* nameBase, Args... args)
{
  TString sFullName = EStoringMode::kSUBDIR == fStoringMode ? Form("tables/%s", nameBase) : nameBase;
  auto* pTable      = MakeObject<CbmQaTable>(sFullName, args...);
  return pTable;
}

#endif  // CbmQaIO_h
