/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmQaTask.h
/// \brief  A base class for CBM QA task logic
/// \author S.Zharko <s.zharko@gsi.de>
/// \data   12.01.2023

#ifndef CbmQaTask_h
#define CbmQaTask_h 1

#include "CbmQaIO.h"
#include "CbmQaTable.h"

#include "FairTask.h"
#include "Logger.h"

#include "TCanvas.h"
#include "TEfficiency.h"
#include "TFolder.h"
#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TParameter.h"
#include "TProfile.h"
#include "TProfile2D.h"
#include "TProfile3D.h"
#include "TROOT.h"

#include <algorithm>
#include <regex>
#include <string_view>
#include <type_traits>

#include <yaml-cpp/yaml.h>

/// Class CbmQaTask is to be inherited with a particular QA-task. It provides mechanisms for storage and management
/// of QA canvases and histograms management
class CbmQaTask : public FairTask, public CbmQaIO {
public:
  /// Constructor from parameters
  /// \param  name     Name of the task
  /// \param  prefix   Unique prefix for writeable root objects
  /// \param  verbose  Verbose level
  /// \param  isMCUsed Flag: true - MC information is used, false - only reconstructed data QA is processed
  CbmQaTask(const char* name, const char* prefix, int verbose, bool isMCUsed);

  /// Default constructor
  CbmQaTask() = delete;  // TODO: Let's see, what can happen, if one deletes default constructor

  /// Destructor
  virtual ~CbmQaTask() = default;

  // Copy and move semantics
  CbmQaTask(const CbmQaTask&) = delete;
  CbmQaTask(CbmQaTask&&)      = delete;
  CbmQaTask& operator=(const CbmQaTask&) = delete;
  CbmQaTask& operator=(CbmQaTask&&) = delete;

  /// Returns flag, whether MC information is used or not in the task
  bool IsMCUsed() const { return fbUseMC; }

  /// FairTask: Task initialization in the beginning of the run
  InitStatus Init();

  /// FairTask: Task reinitialization
  InitStatus ReInit();

  /// FairTask: Defines action of the task in the event/TS
  void Exec(Option_t* /*option*/);

  /// FairTask: Defines action of the task in the end of run
  void Finish();

  /// @brief Sets task configuration file
  /// @param filename  Name of file
  ///
  /// Sets name of YAML configuration file, which defines parameters needed for the task. The file is assumed
  /// to contain root branch "qa/[task name]". A subbranch "histograms" contains a list of histograms with pre defined
  /// parameters: title, number of bins and axis ranges.
  void SetConfigName(const char* filename) { fsConfigName = filename; }

  ClassDef(CbmQaTask, 0);

protected:
  // *****************************************************
  // ** Functions accessible inside the derived classes **
  // *****************************************************

  /// \brief Method to check, if the QA results are acceptable
  virtual bool Check() = 0;

  /// De-initialize the task
  virtual void DeInit() {}

  /// Initializes data branches
  virtual InitStatus InitDataBranches() { return kSUCCESS; }

  /// Initializes histograms
  virtual InitStatus InitHistograms() { return kSUCCESS; }

  /// Initializes canvases
  virtual InitStatus InitCanvases() { return kSUCCESS; }

  /// Initializes event / time-slice
  virtual InitStatus InitTimeSlice() { return kSUCCESS; }

  /// Method to fill histograms per event or time-slice
  virtual void FillHistograms() {}

  /// Get current event number
  int GetEventNumber() const { return fNofEvents.GetVal(); }


  // ***********************
  // ** Utility functions **
  // ***********************

  /// Checks range of variable
  /// \param name  Name of the variable
  /// \param var   Variable to check
  /// \param lo    Lower limit of the variable
  /// \param hi    Upper limit of the variable
  /// \return  False, if variable exits the range
  template<typename T>
  bool CheckRange(std::string_view name, const T& var, const T& lo, const T& hi) const;


private:
  /// @brief De-initializes this task
  void DeInitBase();

  bool fbUseMC     = false;  ///< Flag, if MC is used

  TParameter<int> fNofEvents {"nEvents", 0};  ///< Number of processed events

  TString fsConfigName      = "";       ///< Name of YAML configuration file
  YAML::Node* fpCurrentNode = nullptr;  ///< Pointer to current node of the configuration file
};


// *************************************************
// ** Inline and template function implementation **
// *************************************************

// ---------------------------------------------------------------------------------------------------------------------
//
template<typename T>
bool CbmQaTask::CheckRange(std::string_view name, const T& var, const T& lo, const T& hi) const
{
  if (std::clamp(var, lo, hi) == lo) {
    LOG(error) << fName << ": " << name << " is found to be under the lower limit (" << var << " < " << lo << ')';
    return false;
  }
  if (std::clamp(var, lo, hi) == hi) {
    LOG(error) << fName << ": " << name << " is found to be above the upper limit (" << var << " > " << hi << ')';
    return false;
  }
  return true;
}

#endif  // CbmQaTask_h
